#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include <time.h>
#include <stddef.h>
#include <errno.h>

#define progver "%s: scan/change symbolic links - V0.4 - by Mark Lord\n\n"
static char *progname;
static int verbose = 0, fix_links = 0, recurse = 0;

void fix_symlink (path, my_dev)
char *path;
dev_t my_dev;
{
	static char lpath[PATH_MAX], new_lpath[PATH_MAX];
	char *p, *np, *lp, *tail;
	struct stat stbuf;
	int c;

	if ((c = readlink(path, lpath, sizeof(lpath))) == -1) {
		fprintf(stderr, "could not access symlink: %s\n", path); 
		return;
	}

	lpath[c] = '\0';
	new_lpath[0] = '\0';

	if (lpath[0] != '/') {
		strcat(new_lpath,path);
		if (((c = strlen(new_lpath)) > 0) && (new_lpath[c-1] == '/'))
			new_lpath[c-1] = '\0';
		if ((p = strrchr(new_lpath,'/')) != NULL)
			*p = '\0';
		strcat(new_lpath,"/");
	}
	strcat(new_lpath,lpath);
	if (stat(new_lpath, &stbuf) == -1) {
		printf("dangling: %s -> %s\n", path, lpath);
		return;
	}
	if (stbuf.st_dev != my_dev) {
		printf("other_fs: %s -> %s\n", path, lpath);
		return;
	}
	if (lpath[0] != '/') {
		if (verbose)
			printf("relative: %s -> %s\n", path, lpath);
		return;
	}
	printf("absolute: %s -> %s\n", path, lpath);
	if (!fix_links) {
		return;
	}

	if (unlink (path)) {
		fprintf(stderr, "could not unlink: %s -> %s, errno=%d\n", path, lpath, errno);
		printf("ERROR:    %s -> %s\n", path, lpath);
		return;
	}

	p = path;
	tail = lp = lpath;
	while (*p && (*p == *lp)) {
		if (*p++ == '/') {
			while (*p == '/') ++p;
			while (*lp == '/') tail = ++lp;
		} else
			++lp;
	}

	np = new_lpath;
	while (*p) {
		if (*p++ == '/') {
			*np++ = '.';
			*np++ = '.';
			*np++ = '/';
			while (*p == '/') ++p;
		}
	}
	strcpy (np, tail);

	if (symlink(new_lpath, path)) {
		fprintf(stderr, "could not make link: %s -> %s, errno=%d\n", path, new_lpath, errno);
		printf("ERROR:    %s -> %s\n", path, lpath);
	} else {
		printf("changed:  %s -> %s\n", path, new_lpath);
	}
}

void dirwalk (path, pathlen)
char *path;
int pathlen;
{
 	char *np;
	DIR *dfd;
	dev_t my_dev;
	static struct stat stbuf;
	static struct direct *dp;

	if (lstat(path, &stbuf) == -1) {
		fprintf(stderr, "could not access dir: %s\n", path); 
		return;
	}
	my_dev = stbuf.st_dev;

	if ((dfd = opendir(path)) == NULL) {
		fprintf(stderr, "could not access dir: %s\n", path);
		return;
	}

	np = path + pathlen;
	if (*(np-1) != '/') *np++ = '/'; 
	while ((dp = readdir(dfd)) != NULL ) {
		strcpy(np, dp->d_name);
                if (strcmp(np, ".") && strcmp(np,"..")) {
			if (lstat(path, &stbuf) == -1) {
				fprintf(stderr, "could not access file: %s\n", path); 
			} else if (stbuf.st_dev == my_dev) {
				if (S_ISLNK(stbuf.st_mode)) {
					fix_symlink (path, my_dev);
				} else if (recurse && S_ISDIR(stbuf.st_mode)) {
					dirwalk(path, strlen(path));
				}
			}
		}
	} 
	closedir(dfd);
	path[pathlen] = '\0';
}

void usage_error ()
{
	fprintf(stderr, progver, progname);
	fprintf(stderr, "Usage:\t%s [-crv] dirlist\n\n", progname);
	fprintf(stderr, "Flags:\t-c == change absolute links to relative\n\t-r == recurse into subdirs\n\t-v == verbose (show all symlinks)\n\n");
	exit(1);
}

main(argc, argv)
int argc;
char **argv;
{
	static char path[PATH_MAX];
	int dircount = 0;
	char c, *p;

	if  ((progname = (char *) strrchr(*argv, '/')) == NULL)
                progname = *argv;
        else
                progname++;

	while (--argc) {
		p = *++argv;
		if (*p == '-') {
			if (*++p == '\0')
				usage_error();
			while (c = *p++) {
				if (c == 'v')		verbose   = 1;
				else if (c == 'c')	fix_links = 1;
				else if (c == 'r')	recurse   = 1;
				else			usage_error();
			}
		} else {
			strcpy(path, p);
			dirwalk(path, strlen(path));
			++dircount;
		}
	}
	if (dircount == 0)
		usage_error();
	exit (0);
}
